"""Base settings for a Django Rest Framework Project."""
import os
from t3.env import load_env
from t3.django.util import get_base_dir


# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = get_base_dir()


# Get env dictionary.  This feeds from env, .env, and then vcap services
# For local dev, move .env.example to .env and play with settings
# In CI, a service.yaml is likely to be used
_env = load_env(dot_env_path=BASE_DIR)


# Create of merge INSTALLED_APPS
# print(globals().get('INSTALLED_APPS', []))
# INSTALLED_APPS = globals().get('INSTALLED_APPS', [])
# print(globals().get('INSTALLED_APPS', []))
DRF_INSTALLED_APPS = [
    'django_filters',
    'crispy_forms',
    'rest_framework',

    'drf_yasg',
]


# Rest Framework Settings
REST_FRAMEWORK = {
    'DEFAULT_RENDERER_CLASSES': (
        'djangorestframework_camel_case.render.CamelCaseJSONRenderer',
    ),
    'DEFAULT_PARSER_CLASSES': (
        'djangorestframework_camel_case.parser.CamelCaseJSONParser',
    ),
    'DEFAULT_AUTHENTICATION_CLASSES': (),
    'DEFAULT_PAGINATION_CLASS': 'rest_framework.pagination.LimitOffsetPagination',
    'EXCEPTION_HANDLER': 't3.drf.views.enveloped_exception_handler',
    'PAGE_SIZE': 25,
}

DOCS_TITLE = os.getenv('DOCS_TITLE', 'T3 API')
DOCS_VERSION = os.getenv('DOCS_VERSION', 'v1')



from contextlib import suppress
from rest_framework import views
from rest_framework import viewsets
# from rest_framework import generics

from django.utils.decorators import method_decorator
from drf_yasg.utils import swagger_auto_schema

from t3.drf_adapters import mixins


class ResourceViewSetBase(viewsets.ViewSetMixin, views.APIView):
    resource_class = None
    serializer_class = None

    lookup_field = 'pk'
    lookup_url_kwarg = None

    def get_queryset(self):
        return self.get_resource(**self.kwargs)

    def get_object(self):
        # TODO: Add filter
        # TODO: Add permissions
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        assert lookup_url_kwarg in self.kwargs, (
            'Expected view %s to be called with a URL keyword argument '
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            'attribute on the view correctly.' %
            (self.__class__.__name__, lookup_url_kwarg)
        )
        # resource = self.get_resource_class()
        key = self.kwargs[lookup_url_kwarg]
        return self.get_resource(key, **self.kwargs)

    def get_resource_class(self):
        assert self.resource_class is not None, '`resource_class` missing'
        return self.resource_class
        # return resource().retrieve(key)

    def get_resource(self, *args, **kwargs):
        # assert self.resource_class is not None, '`resource_class` missing'
        resource_class = self.get_resource_class()
        kwargs['request'] = self.request
        return resource_class(*args, **kwargs)

    def get_serializer_class(self):
        assert self.serializer_class is not None, '`serializer_class` missing'
        return self.serializer_class

    def get_serializer(self, *args, **kwargs):
        serializer_class = self.get_serializer_class()
        kwargs['context'] = self.get_serializer_context()
        return serializer_class(*args, **kwargs)

    def get_serializer_context(self):
        return {
            'request': self.request,
            'format': self.format_kwarg,
            'view': self
        }


class ResourceViewSet(ResourceViewSetBase):
    @classmethod
    def factory(cls, resource_class, serializer_class):
        # Add mixins based on available resource verbs
        bases = [cls, ]

        if hasattr(resource_class, 'list'):
            bases.append(mixins.ListResourceMixin)
        if hasattr(resource_class, 'create'):
            bases.append(mixins.CreateResourceMixin)
        if hasattr(resource_class, 'retrieve'):
            bases.append(mixins.RetrieveResourceMixin)
        if hasattr(resource_class, 'update'):
            bases.append(mixins.UpdateResourceMixin)
        if hasattr(resource_class, 'destroy'):
            bases.append(mixins.DestroyResourceMixin)

        # Create a new class so we can set the resource and serializer
        # without changing the primary class
        bases = tuple(bases)
        new_cls = type(cls.__name__, bases, {})

        new_cls.resource_class = resource_class
        # Create a new serializer class so we can set the
        # resource on it and not effect the primary serializer
        new_serializer_class = type(serializer_class.__name__, (serializer_class, ), {})
        new_serializer_class.resource_class = resource_class
        new_cls.serializer_class = new_serializer_class

        return new_cls



from collections import OrderedDict
from contextlib import suppress
from t3 import log
from t3.util import get_response_type


class DjangoFormatterMixin(log.StandardFormatter):
    def process(self, record):
        out = super().process(record)

        with suppress(AttributeError):
            # Put request id to top of log envelope
            out['request_id'] = record.request_id
            out.move_to_end('request_id', last=False)

        with suppress(AttributeError):
            # Add response info
            out['response'] = OrderedDict([
                ('status_code', record.status_code),
                ('response_type', get_response_type(record.status_code)),
                ('remote_address', record.request.getpeername()[0]),
                ('local_address', record.request.getsockname()[0]),
            ])

        return out


class DjangoFormatter(DjangoFormatterMixin, log.StandardFormatter):
    pass


class DjangoJsonFormatter(DjangoFormatterMixin, log.JsonFormatter):
    pass


class DjangoJsonPrettyFormatter(DjangoFormatterMixin, log.JsonPrettyFormatter):
    pass



from os import environ
from os.path import join
from functools import lru_cache

from dotenv import load_dotenv


@lru_cache()
def load_env(dot_env_path, filename='.env'):
    load_dotenv(dotenv_path=join(dot_env_path, filename))
    return environ



from rest_framework import serializers


class ResourceSerializer(serializers.Serializer):
    resource_class = None

    def create(self, validated_data):
        assert self.resource_class is not None, '`resource_class` missing'
        resource = self.resource_class()
        return resource.create(**validated_data)

    def update(self, instance, validated_data):
        if self.partial:
            return instance.partial_update(**validated_data)
        return instance.update(**validated_data)

    def partial_update(self, instance, validated_data):
        return instance.partial_update(**validated_data)


class AccountSerializer(ResourceSerializer):
    id = serializers.UUIDField(read_only=True)
    phone = serializers.CharField()
    email = serializers.EmailField()
    name = serializers.CharField()
    points = serializers.IntegerField(read_only=True)
    meta = serializers.JSONField()



"""Basic Django Models."""
from django.db import models
from django_prometheus.models import model_inserts, model_updates, model_deletes


class PrometheusModelBase(object):
    # Force create the labels for this model in the counters. This
    # is not necessary but it avoids gaps in the aggregated data.
    def __init__(self, *args, **kwargs):
        self.model_name = self._meta.db_table
        model_inserts.labels(self.model_name)
        model_updates.labels(self.model_name)
        model_deletes.labels(self.model_name)
        super().__init__(*args, **kwargs)

    def _do_insert(self, *args, **kwargs):
        model_inserts.labels(self.model_name).inc()
        return super()._do_insert(*args, **kwargs)

    def _do_update(self, *args, **kwargs):
        model_updates.labels(self.model_name).inc()
        return super()._do_update(*args, **kwargs)

    def delete(self, *args, **kwargs):
        model_deletes.labels(self.model_name).inc()
        return super().delete(*args, **kwargs)


class T3Model(PrometheusModelBase, models.Model):
    """Basic Django Model that exports monitoring metrics for Prometheus."""

    class Meta:
        abstract = True



# pylint: disable=W0223
from django.conf import settings
from storages.backends.s3boto3 import S3Boto3Storage


class MediaStorage(S3Boto3Storage):
    location = settings.MEDIAFILES_LOCATION
    bucket_name = settings.MEDIAFILES_BUCKET
    default_acl = 'private'


class StaticStorage(S3Boto3Storage):
    location = settings.STATICFILES_LOCATION
    bucket_name = settings.STATICFILES_BUCKET
    default_acl = 'public-read'



from django.urls import path, re_path
from django.conf import settings
from rest_framework import permissions
from rest_framework.documentation import include_docs_urls
from drf_yasg.views import get_schema_view
from drf_yasg import openapi


def build_doc_urls(version=None):
    if not version:
        version = settings.DOCS_VERSION

    # Swagger
    schema_view = get_schema_view(
        openapi.Info(
            title=settings.DOCS_TITLE,
            default_version=version,
        ),
        validators=['flex', 'ssv'],
        public=True,
        permission_classes=(permissions.AllowAny,),
    )

    return [
        # Vanilla DRF Docs
        path('docs/', include_docs_urls(title=settings.DOCS_TITLE)),

        # Swagger Docs
        re_path(r'^swagger(?P<format>\.json|\.yaml)$', schema_view.without_ui(cache_timeout=None), name='schema-json'),
        path('swagger/', schema_view.with_ui('swagger', cache_timeout=None), name='schema-swagger-ui'),
        path('redoc/', schema_view.with_ui('redoc', cache_timeout=None), name='schema-redoc'),
    ]
